#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from hakkero.dataset.dataloader import PackPadLoader
from hakkero.dataset.dataloader import PadLoader
from hakkero.dataset.dataloader import PreferencePadLoader
from hakkero.dataset.dataloader import PreferenceUnpadLoader
from hakkero.dataset.dataloader import UnpadLoader
from hakkero.dataset.image import process_messages
from hakkero.dataset.mixed_dataset import get_dataset
from hakkero.dataset.utils import IGNORE_INDEX


def get_data(
    config,
    dp_rank,
    dp_world_size,
    tokenizer,
    batch_size,
    max_length,
    num_epochs=10,
    homogeneous=False,
    seed=9527,
    n_workers=2,
    use_unpad_data=False,
    use_unpad_in_pad=False,
    use_cu_seqlens=False,
    st_segment=None,
    st_tokenize=None,
    pad_with_ignore_index=False,
    add_bos_token=True,
    add_eos_token=True,
    norm_weight_with_n_targets=False,
    keep_think="no",
    processor=None,
    **kwargs
):
    dataset = get_dataset(
        config=config,
        tokenizer=tokenizer,
        processor=processor,
        num_epochs=num_epochs,
        max_length=max_length,
        homogeneous=homogeneous,
        seed=seed,
        rank=dp_rank,
        world_size=dp_world_size,
        n_workers=n_workers,
        # segment & tokenize strategy
        st_segment=st_segment,
        st_tokenize=st_tokenize,
        #
        add_bos_token=add_bos_token,
        add_eos_token=add_eos_token,
        #
        norm_weight_with_n_targets=norm_weight_with_n_targets,
        keep_think=keep_think,
        **kwargs
    )

    is_preference = "preference" in list(config.values())[0].get("strategy", {}).get("st_tokenize", "")
    is_preference = is_preference or (st_tokenize is not None and "preference" in st_tokenize)

    if use_unpad_data:
        if not is_preference:
            loader = UnpadLoader(dataset, max_total_length=batch_size * max_length)
        else:
            loader = PreferenceUnpadLoader(dataset, max_total_length=batch_size * max_length)
    else:
        if pad_with_ignore_index:
            padding_id = IGNORE_INDEX
        elif tokenizer.pad_token_id is None:
            padding_id = tokenizer.eos_token_id
        else:
            padding_id = tokenizer.pad_token_id

        if kwargs.pop("packed", False):
            loader = PackPadLoader(
                dataset,
                batch_size=batch_size,
                padding_id=padding_id,
                unpad=use_unpad_in_pad,
                bos_id=tokenizer.bos_id,
                eos_id=tokenizer.eos_id,
            )
        else:
            if not is_preference:
                loader = PadLoader[is_preference](
                    dataset, batch_size=batch_size, padding_id=padding_id, unpad=use_unpad_in_pad
                )
            else:
                loader = PreferencePadLoader[is_preference](
                    dataset, batch_size=batch_size, padding_id=padding_id, unpad=use_unpad_in_pad
                )

    if use_unpad_data or use_unpad_in_pad:
        if use_cu_seqlens:
            forward_keys = ["input_ids", "cu_seqlens"]
        else:
            forward_keys = ["input_ids", "position_ids"]
    else:
        forward_keys = ["input_ids", "attention_mask"]

    return dataset, loader, forward_keys



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

KEEP_THINK_ALL = "all"
KEEP_THINK_LAST = "last"
KEEP_THINK_NO = "no"
keep_think_strategy = [KEEP_THINK_NO, KEEP_THINK_ALL, KEEP_THINK_LAST]


KEEP_THINK_REMAP = {
    "{% if '</think>' in content %}{% set content = content.split('</think>')[-1] %}{% endif %}": "{% if (not keep_think is defined or not keep_think) and '</think>' in content %}{% set content = content.split('</think>')[-1] %}{% endif %}",
    "{%- set content = message.content.split('</think>')[-1].lstrip('\\n') %}": "{%- set content = message.content %}{% if (not keep_think is defined or not keep_think) and '</think>' in content %}{% set content = message.content.split('</think>')[-1].lstrip('\\n') %}{% endif %}",
}


def change_chat_template_with_keep_think_remap(chat_template, keep_think):
    new_chat_template = chat_template

    if keep_think in {KEEP_THINK_ALL, KEEP_THINK_LAST}:
        for key, value in KEEP_THINK_REMAP.items():
            new_chat_template = new_chat_template.replace(key, value)
    else:
        keep_think = KEEP_THINK_NO

    return new_chat_template, keep_think



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import itertools
from collections import defaultdict
from typing import Iterable

import numpy as np
import torch
import torch.utils.data
from torch.nn.utils.rnn import pad_sequence

from hakkero.dataset.misc import unpack_packed_tokens
from hakkero.dataset.utils import IGNORE_INDEX


def select(seq, indices, reverse=False):
    if not isinstance(indices, set):
        indices = set(indices)

    if reverse:
        return [seq for i, seq in enumerate(seq) if i not in indices]

    return [seq for i, seq in enumerate(seq) if i in indices]


class CudaPrefetcher(Iterable):
    def __init__(self, loader):
        self.loader = iter(loader)
        self.stream = torch.cuda.Stream()
        self.next = None

        self.preload()

    def preload(self):
        try:
            self.next = next(self.loader)
        except StopIteration:
            self.next = None
            return

        with torch.cuda.stream(self.stream):
            for key in self.next.keys():
                if isinstance(self.next[key], torch.Tensor):
                    self.next[key] = self.next[key].cuda(non_blocking=True)

    def __next__(self):
        torch.cuda.current_stream().wait_stream(self.stream)
        current = self.next

        if current is None:
            raise StopIteration()

        for value in current.values():
            if isinstance(value, torch.Tensor):
                value.record_stream(torch.cuda.current_stream())

        self.preload()

        return current

    def __iter__(self):
        return self


class Loader(torch.utils.data.IterableDataset):
    def prefetch(self, num_workers=2, prefetch_factor=20, pin_memory=True, drop_last=False):
        loader = torch.utils.data.DataLoader(
            self,
            batch_size=None,
            collate_fn=lambda x: x,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=pin_memory,
            drop_last=drop_last,
        )
        return CudaPrefetcher(loader)

    @staticmethod
    def get_stats(task_ids, useds, failed):
        """stats:
        {
            "dataset_id": {
                "epoch": {data_id, data_id, ...}
            }
        }
        """
        # task -> epoch -> set(indices)

        stats = dict()
        for task, used in itertools.chain(zip(task_ids, useds), failed):
            if task not in stats:
                stats[task] = defaultdict(set)
            for epoch, index in used:
                stats[task][epoch].add(index)
        return stats


class PadLoaderBase(Loader):
    def __init__(self, dataset, batch_size, padding_id, bos_id=None, eos_id=None, unpad=False):
        assert padding_id is not None, "padding_id should not be None"

        self.bos_id = bos_id
        self.eos_id = eos_id

        self.dataset = dataset
        self.padding_id = padding_id
        self.batch_size = batch_size
        self.unpad = unpad

        self.useds = []
        self.failed = []

        self.input_ids = []
        self.labels = []

        self.pixel_values = []
        self.image_grid_thw = []

        self.lengths = []
        self.n_targets = []
        self.task_ids = []

    def exceed(self):
        return len(self.labels) == self.batch_size

    def __iter__(self):
        for sample in self.dataset:
            self.put(sample)

            if not self.exceed():
                continue
            yield self.pop()

        if len(self.labels):
            yield self.pop()

    def put(self, sample):
        raise NotImplementedError()

    def pop(self):
        raise NotImplementedError()


class PadLoader(PadLoaderBase):
    def put(self, sample):
        if "input" not in sample:
            self.failed.append((sample["task"], sample["used"]))
            return

        self.useds.append(sample["used"])
        self.task_ids.append(sample["task"])

        self.input_ids.append(sample["input"])
        self.labels.append(sample["label"])
        if "pixel_values" in sample:
            self.pixel_values.extend(sample["pixel_values"])
        if "image_grid_thw" in sample:
            self.image_grid_thw.extend(sample["image_grid_thw"])

        self.lengths.append(sample["length"])
        self.n_targets.append(sample["n_target"])

    def pop(self):
        batch = {
            "n_samples": torch.tensor(len(self.lengths), dtype=torch.long),
            "n_tokens": torch.tensor(sum(self.lengths), dtype=torch.long),
            "n_targets": torch.tensor(sum(self.n_targets), dtype=torch.long),
            "stats": self.get_stats(self.task_ids, self.useds, self.failed),
        }

        if not self.unpad:
            batch["input_ids"] = pad_sequence(self.input_ids, batch_first=True, padding_value=self.padding_id).long()
            batch["labels"] = pad_sequence(self.labels, batch_first=True, padding_value=IGNORE_INDEX).long()
            batch["attention_mask"] = batch["input_ids"].ne(self.padding_id).long()
            batch["task_ids"] = torch.stack(
                [torch.full((batch["input_ids"].shape[-1],), task, dtype=torch.long) for task in self.task_ids], dim=0
            )
            if self.pixel_values:
                batch["pixel_values"] = torch.stack(self.pixel_values, dim=0)
            if self.image_grid_thw:
                batch["image_grid_thw"] = torch.stack(self.image_grid_thw, dim=0)
        else:
            batch["input_ids"] = torch.cat(self.input_ids, dim=0).unsqueeze(0)
            batch["labels"] = torch.cat(self.labels, dim=0).unsqueeze(0)
            batch["cu_seqlens"] = torch.tensor([0] + self.lengths).cumsum(dim=-1).int()
            batch["position_ids"] = torch.cat(
                [torch.arange(length, dtype=torch.long) for length in self.lengths], 0
            ).unsqueeze(0)
            batch["task_ids"] = torch.cat(
                [torch.full((l,), t, dtype=torch.long) for t, l in zip(self.task_ids, self.lengths)], dim=0
            )
            if self.pixel_values:
                batch["pixel_values"] = torch.cat(self.pixel_values, dim=0)
            if self.image_grid_thw:
                batch["image_grid_thw"] = torch.cat(self.image_grid_thw, dim=0)

        self.useds = []
        self.failed = []

        self.input_ids = []
        self.labels = []

        self.pixel_values = []
        self.image_grid_thw = []

        self.lengths = []
        self.n_targets = []
        self.task_ids = []

        return batch


class PackPadLoader(PadLoader):
    def pop(self):
        batch = {"stats": self.get_stats(self.task_ids, self.useds, self.failed)}

        if not self.unpad:
            raise NotImplementedError("not support")
        else:
            re_input_ids = []
            re_labels = []
            re_lengths = []
            for cur_input_ids in self.input_ids:
                unpacked = unpack_packed_tokens(cur_input_ids.tolist(), self.bos_id, self.eos_id)
                re_input_ids.extend([torch.tensor(s[:-1], dtype=torch.long) for s in unpacked])

                re_labels.extend([torch.tensor(s[1:], dtype=torch.long) for s in unpacked])

                cur_lengths = [len(s[1:]) for s in unpacked]
                re_lengths.extend(cur_lengths)

            batch["input_ids"] = torch.cat(re_input_ids, dim=0).unsqueeze(0)
            batch["labels"] = torch.cat(re_labels, dim=0).unsqueeze(0)
            batch["cu_seqlens"] = torch.tensor([0] + re_lengths).cumsum(dim=-1).int()
            batch["position_ids"] = torch.cat(
                [torch.arange(length, dtype=torch.long) for length in re_lengths], dim=0
            ).unsqueeze(0)

            batch["n_tokens"] = torch.tensor(sum(re_lengths), dtype=torch.long)
            batch["n_samples"] = torch.tensor(len(re_lengths), dtype=torch.long)

        self.useds = []
        self.failed = []

        self.input_ids = []
        self.labels = []

        self.pixel_values = []
        self.image_grid_thw = []

        self.lengths = []
        self.n_targets = []
        self.task_ids = []

        return batch


class PreferencePadLoader(PadLoaderBase):
    def put(self, sample):
        if "inputs" not in sample:
            self.failed.append((sample["task"], sample["used"]))
            return

        self.useds.append(sample["used"])

        self.input_ids.append(sample["inputs"])
        self.labels.append(sample["labels"])

        self.lengths.append(sample["length"])
        self.n_targets.append(sample["n_target"])
        self.task_ids.append(sample["task"])

    def pop(self):
        batch = {
            "input_ids": dict(),
            "labels": dict(),
            "task_ids": dict(),
            "n_samples": torch.tensor(len(self.lengths), dtype=torch.long),
            "n_tokens": torch.tensor(
                sum(pl for part_lengths in self.lengths for pl in part_lengths.values()), dtype=torch.long
            ),
            "n_targets": torch.tensor(
                sum(pt for part_n_targets in self.lengths for pt in part_n_targets.values()), dtype=torch.long
            ),
            "stats": self.get_stats(self.task_ids, self.useds, self.failed),
        }

        if not self.unpad:
            batch["attention_mask"] = dict()

            for key in ("chosen", "rejected"):
                batch["input_ids"][key] = pad_sequence(
                    [d[key] for d in self.input_ids], batch_first=True, padding_value=self.padding_id
                ).long()
                batch["labels"][key] = pad_sequence(
                    [d[key] for d in self.labels], batch_first=True, padding_value=IGNORE_INDEX
                ).long()

                batch["attention_mask"][key] = batch["input_ids"][key].ne(self.padding_id).long()

                batch_length = batch["input_ids"][key].shape[-1]
                batch["task_ids"][key] = torch.stack(
                    [torch.full((batch_length,), task, dtype=torch.long) for task in self.task_ids], dim=0
                )
        else:
            batch["cu_seqlens"] = dict()
            batch["max_seqlen"] = dict()
            batch["position_ids"] = dict()

            for key in ("chosen", "rejected"):
                batch["input_ids"][key] = torch.cat([d[key] for d in self.input_ids], 0).unsqueeze(0)
                batch["labels"][key] = torch.cat([d[key] for d in self.labels], 0).unsqueeze(0)

                seqlens = [p[key] for p in self.lengths]
                batch["cu_seqlens"][key] = torch.tensor([0] + seqlens).cumsum(dim=-1).int()

                batch["max_seqlen"][key] = max(seqlens)

                batch["position_ids"][key] = torch.cat(
                    [torch.arange(l, dtype=torch.long) for l in seqlens], 0
                ).unsqueeze(0)

                batch["task_ids"][key] = torch.cat(
                    [torch.full((l,), t, dtype=torch.long) for t, l in zip(self.task_ids, seqlens)], 0
                ).unsqueeze(0)

        self.useds = []
        self.failed = []

        self.input_ids = []
        self.labels = []

        self.lengths = []
        self.n_targets = []
        self.task_ids = []

        return batch


class UnpadLoaderBase(Loader):
    def __init__(self, dataset, max_total_length, n_cache=3, max_explore=32):
        self.dataset = dataset
        self.max_total_length = max_total_length
        self.n_cache = n_cache
        self.max_explore = max_explore  # maximum comparison times for find_batch_combination
        self.total_length = 0

        self.useds = []
        self.failed = []

        self.lengths = []
        self.input_ids = []
        self.labels = []

        self.n_targets = []
        self.task_ids = []

        self.seqlens = []

    def find_batch_combination(self):
        # find the longest batch combination that is smaller than max_total_length
        if not self.lengths:
            return []

        if min(self.lengths) > self.max_total_length:
            raise ValueError(
                f"No the minimal segment length {min(self.lengths)} does not fit into max_total_length {self.max_total_length}"
            )

        # simple return all samples
        if self.total_length < self.max_total_length:
            return list(range(len(self.lengths)))

        # arrange the indices for better backtrack heuristics:
        # 1. fill the batch with random samples
        # 2. modify the batch by swapping the shorter samples
        # the batch is not necessarily biased as we are swapping shorter samples with shorter ones.
        i = 0
        for i, length_sum in enumerate(itertools.accumulate(self.lengths)):
            if length_sum > self.max_total_length:
                break

        bt_indices = np.argsort(self.lengths[:i])[::-1].tolist() + (np.argsort(self.lengths[i:]) + i).tolist()

        num_explore = 0
        best_indices = []
        best_total_length = 0
        stack = [(0, 0, [])]

        # depth-first backtracking with early stopping
        # the default combination will always be considered first
        # the try to modify the last elements of the combination
        while stack:
            bt_index, current_total_length, current_batch = stack.pop()
            if current_total_length > self.max_total_length:
                # skip as future combination will certainly exceed max_total_length
                continue
            elif current_total_length > best_total_length:
                # improved
                best_total_length = current_total_length
                best_indices = current_batch[:]
            else:
                # fail to improve
                num_explore += 1

            # early stop to avoid too much comparison
            if num_explore > self.max_explore:
                break

            # stop if we've considered all samples
            if bt_index == len(self.lengths):
                break

            # exclude the current sample and explore further
            stack.append((bt_index + 1, current_total_length, current_batch))

            # include the current sample and explore further
            stack.append(
                (
                    bt_index + 1,
                    current_total_length + self.lengths[bt_indices[bt_index]],
                    current_batch + [bt_indices[bt_index]],
                )
            )

        return best_indices

    def __iter__(self):
        for sample in self.dataset:
            self.put(sample)
            if self.exceed():
                yield self.pop()
        while self.input_ids:
            yield self.pop()

    def exceed(self):
        return self.total_length > self.n_cache * self.max_total_length

    def put(self, sample):
        raise NotImplementedError()

    def pop(self):
        raise NotImplementedError()


class UnpadLoader(UnpadLoaderBase):
    def put(self, sample):
        if "input" not in sample:
            self.failed.append((sample["task"], sample["used"]))
            return

        self.task_ids.append(sample["task"])
        self.useds.append(sample["used"])

        self.lengths.append(sample["length"])
        self.n_targets.append(sample["n_target"])
        self.input_ids.append(sample["input"])
        self.labels.append(sample["label"])
        self.total_length += sample["length"]

    def pop(self):
        indices = self.find_batch_combination()
        lengths = select(self.lengths, indices)
        n_targets = select(self.n_targets, indices)
        task_ids = select(self.task_ids, indices)
        useds = select(self.useds, indices)

        input_ids = select(self.input_ids, indices)
        labels = select(self.labels, indices)

        batch = {
            "input_ids": torch.cat(input_ids, 0).unsqueeze(0),
            "labels": torch.cat(labels, 0).unsqueeze(0),
            "n_samples": torch.tensor(len(lengths), dtype=torch.long),
            "n_tokens": torch.tensor(sum(lengths), dtype=torch.long),
            "n_targets": torch.tensor(sum(n_targets), dtype=torch.long),
            "stats": self.get_stats(task_ids, useds, self.failed),
            "task_ids": torch.cat(
                [torch.full((l,), t, dtype=torch.long) for t, l in zip(task_ids, lengths)], dim=0
            ).unsqueeze(0),
            "position_ids": torch.cat([torch.arange(length, dtype=torch.long) for length in lengths], 0).unsqueeze(0),
            "cu_seqlens": torch.tensor([0] + lengths).cumsum(dim=-1).int(),
            "max_seqlen": max(lengths),
        }

        self.lengths = select(self.lengths, indices, reverse=True)
        self.n_targets = select(self.n_targets, indices, reverse=True)
        self.task_ids = select(self.task_ids, indices, reverse=True)
        self.useds = select(self.useds, indices, reverse=True)
        self.total_length = sum(self.lengths)

        self.input_ids = select(self.input_ids, indices, reverse=True)
        self.labels = select(self.labels, indices, reverse=True)

        self.failed = []

        return batch


class PreferenceUnpadLoader(UnpadLoaderBase):
    def put(self, sample):
        if "inputs" not in sample:
            self.failed.append((sample["task"], sample["used"]))
            return

        length = sum(sample["length"].values())
        self.lengths.append(length)
        self.total_length += length
        n_target = sum(sample["n_target"].values())
        self.n_targets.append(n_target)

        self.task_ids.append(sample["task"])
        self.useds.append(sample["used"])
        self.input_ids.append(sample["inputs"])
        self.labels.append(sample["labels"])

        self.seqlens.append(sample["length"])

    def pop(self):
        indices = self.find_batch_combination()

        lengths = select(self.lengths, indices)
        n_targets = select(self.n_targets, indices)
        task_ids = select(self.task_ids, indices)
        useds = select(self.useds, indices)

        part_input_ids = select(self.input_ids, indices)
        part_labels = select(self.labels, indices)
        part_seqlens = select(self.seqlens, indices)

        # task -> epoch -> set(indices)
        stats = dict()
        for task, used in itertools.chain(zip(task_ids, useds), self.failed):
            if task not in stats:
                stats[task] = defaultdict(set)
            for epoch, index in used:
                stats[task][epoch].add(index)

        batch = {
            "input_ids": dict(),
            "labels": dict(),
            "task_ids": dict(),
            "sample_ids": dict(),
            "position_ids": dict(),
            "cu_seqlens": dict(),
            "max_seqlen": dict(),
            "n_samples": torch.tensor(len(lengths), dtype=torch.long),
            "n_tokens": torch.tensor(sum(lengths), dtype=torch.long),
            "n_targets": torch.tensor(sum(n_targets), dtype=torch.long),
            "stats": self.get_stats(self.task_ids, self.useds, self.failed),
        }

        for key in ("chosen", "rejected"):
            batch["input_ids"][key] = torch.cat([d[key] for d in part_input_ids], 0).unsqueeze(0)
            batch["labels"][key] = torch.cat([d[key] for d in part_labels], 0).unsqueeze(0)

            seqlens = [part_seqlen[key] for part_seqlen in part_seqlens]
            batch["task_ids"][key] = torch.cat(
                [torch.full((l,), t, dtype=torch.long) for t, l in zip(task_ids, seqlens)], 0
            ).unsqueeze(0)
            batch["sample_ids"][key] = torch.cat(
                [torch.full((l,), s, dtype=torch.long) for s, l in enumerate(seqlens)], 0
            ).unsqueeze(0)
            batch["position_ids"][key] = torch.cat([torch.arange(l, dtype=torch.long) for l in seqlens], 0).unsqueeze(0)
            batch["cu_seqlens"][key] = (torch.tensor([0] + seqlens).cumsum(dim=-1).int(),)
            batch["max_seqlen"][key] = max(seqlens)

        self.lengths = select(self.lengths, indices, reverse=True)
        self.n_targets = select(self.n_targets, indices, reverse=True)
        self.task_ids = select(self.task_ids, indices, reverse=True)
        self.useds = select(self.useds, indices, reverse=True)
        self.total_length = sum(self.lengths)

        self.input_ids = select(self.input_ids, indices, reverse=True)
        self.labels = select(self.labels, indices, reverse=True)
        self.seqlens = select(self.seqlens, indices, reverse=True)

        self.failed = []
        return batch



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#


from hakkero.dataset.strategy.errors import SegmentationError
from hakkero.dataset.strategy.errors import TokenizationError
from hakkero.dataset.strategy.keep_thinks import change_chat_template_with_keep_think_remap
from hakkero.dataset.strategy.keep_thinks import KEEP_THINK_ALL
from hakkero.dataset.strategy.keep_thinks import KEEP_THINK_LAST
from hakkero.dataset.strategy.keep_thinks import KEEP_THINK_NO
from hakkero.dataset.strategy.segmentation import concat
from hakkero.dataset.strategy.segmentation import integrous
from hakkero.dataset.strategy.segmentation import naive
from hakkero.dataset.strategy.segmentation import unbiased
from hakkero.dataset.strategy.tokenization import chatml_message
from hakkero.dataset.strategy.tokenization import chatml_preference
from hakkero.dataset.strategy.tokenization import chatml_qwen2_vl_message
from hakkero.dataset.strategy.tokenization import huggingface_message
from hakkero.dataset.strategy.tokenization import huggingface_preference
from hakkero.dataset.strategy.tokenization import legacy

ST_INTEGROUS = "integrous"
ST_CONCAT = "concat"
ST_NAIVE = "naive"
ST_UNBIASED = "unbiased"

segment = {
    ST_INTEGROUS: integrous,
    ST_CONCAT: concat,
    ST_NAIVE: naive,
    ST_UNBIASED: unbiased,
}


ST_LEGACY = "legacy"
ST_HG = "hg"
ST_HG_PREFERENCE = "hg_preference"
ST_CHATML = "chatml"
ST_CHATML_PREFERENCE = "chatml_preference"
ST_CHATML_QWEN_VL = "chatml_qwen2_vl_message"

tokenize = {
    ST_LEGACY: legacy,
    ST_HG: huggingface_message,
    ST_HG_PREFERENCE: huggingface_preference,
    ST_CHATML: chatml_message,
    ST_CHATML_PREFERENCE: chatml_preference,
    ST_CHATML_QWEN_VL: chatml_qwen2_vl_message,
}


ST_SEGMENT = "st_segment"
ST_TOKENIZE = "st_tokenize"
default_strategy = {ST_SEGMENT: ST_NAIVE, ST_TOKENIZE: ST_LEGACY}


def check_tokenizer(tokenizer, st_tokenize, keep_think):
    if st_tokenize in {ST_HG, ST_HG_PREFERENCE}:
        assert hasattr(tokenizer, "apply_chat_template"), "tokenizer should have apply_chat_template"

        assert tokenizer.apply_chat_template(
            [{"role": "user", "content": "test"}], add_generation_prompt=True
        ) != tokenizer.apply_chat_template(
            [{"role": "user", "content": "test"}], add_generation_prompt=False
        ), "add_generation_prompt does not take effect, please modify tokenizer.chat_template"

    old_chat_template = tokenizer.chat_template
    new_chat_template, keep_think = change_chat_template_with_keep_think_remap(old_chat_template, keep_think)

    if keep_think != KEEP_THINK_NO:
        assert (
            old_chat_template != new_chat_template
        ), f"change chat template failed with keep_think='{keep_think}', please modify tokenizer.chat_template, otherwise change keep_think to 'no'"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import re
from typing import List


def unpack_packed_string(packed: str):
    """
    [BOS]abc[EOS][BOS]123[BOS]456[EOS]789 => ['[BOS]abc[EOS]', '[BOS]123', '[BOS]456[EOS]', '789']
    """
    tokens = re.split(r"(\[BOS]|\[EOS])", packed)

    samples = []
    cur = ""

    for tok in tokens:
        if tok == "[BOS]":
            if cur:
                samples.append(cur)
            cur = tok
        elif tok == "[EOS]":
            cur += tok
            samples.append(cur)
            cur = ""
        else:
            cur += tok

    if cur:
        samples.append(cur)

    return samples


def unpack_packed_tokens(packed: List[int], bos_id: int, eos_id: int) -> List[List[int]]:
    samples: List[List[int]] = []
    cur: List[int] = []

    for tid in packed:
        if tid == bos_id:
            if cur:
                samples.append(cur)
            cur = [bos_id]
        elif tid == eos_id:
            cur.append(eos_id)
            samples.append(cur)
            cur = []
        else:
            cur.append(tid)

    if cur:
        samples.append(cur)

    return samples



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from transformers import AutoTokenizer

from hakkero.dataset.mixed_dataset import get_dataset
from hakkero.dataset import PackPadLoader

if __name__ == "__main__":
    config = {
        "pt": {
            "group": "en",
            "name": "pt",
            "epoch": 1,
            "path": f"/Users/qinluo/work/repo/hakkero-dataloader/tests/data/pt.packed",
            "strategy": {"st_segment": "naive", "st_tokenize": "legacy"},
            "weight": 1.0,
        }
    }

    model_path = "/Users/qinluo/work/models/moonshotai/Moonlight-16B-A3B"
    tokenizer = AutoTokenizer.from_pretrained(model_path, trust_remote_code=True)

    max_length = 4096
    batch_size = 2
    n_workers = 2
    shift = False

    ds = get_dataset(
        config,
        tokenizer,
        num_epochs=1,
        max_length=max_length,
        homogeneous=True,
        seed=-1,
        rank=0,
        world_size=1,
        n_workers=n_workers,
        shift=False,
    )

    dataloader = PackPadLoader(
        ds, batch_size=batch_size, padding_id=tokenizer.pad_token_id, unpad=True, bos_id=tokenizer.bos_id, eos_id=tokenizer.eos_id
    )

    total_samples = 0
    for step, batch in enumerate(dataloader, start=1):
        total_samples += batch["n_samples"]
        print(f"step={step}, samples={batch['n_samples']}, n_tokens={batch['n_tokens']}, total_samples={total_samples}")
        ds.track(batch["stats"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import glob
import hashlib
import json
import os.path
import shutil
from pathlib import Path

import numpy as np


def object_to_dict(obj):
    return dict((key, value) for key, value in obj.__dict__.items() if not callable(value) and not key.startswith("_"))


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (set, tuple)):
            return list(obj)

        if isinstance(obj, np.integer):
            return int(obj)

        if isinstance(obj, np.floating):
            return float(obj)

        if isinstance(obj, np.ndarray):
            return obj.tolist()

        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            return object_to_dict(obj)


def load_json(filename):
    with open(filename, "r", encoding="utf-8") as fin:
        return json.load(fin)


def save_json(data, filename):
    with open(filename, "w") as fd:
        json.dump(data, fd, ensure_ascii=False, indent=2)


def load_jsonl(filename):
    paths = filename
    if isinstance(paths, str):
        paths = [paths]

    files = []
    for path in paths:
        if os.path.isdir(path):
            files.extend([os.path.join(dp, f) for dp, _, fn in os.walk(os.path.expanduser(path)) for f in fn])
        else:
            files.extend([path])

    data = []
    for filename in files:
        with open(filename, "r", encoding="utf-8") as fin:
            data.extend([json.loads(line) for line in fin])

    return data


def save_jsonl(data, filename):
    with open(filename, "w") as fd:
        for line in data:
            fd.write(json.dumps(line, ensure_ascii=False) + "\n")


def special_copy(src, dst, exts=None):
    if src is None or dst is None or not os.path.exists(src):
        return

    if exts is None:
        exts = []

    filenames = []
    for ext in exts:
        filenames.extend(glob.glob(os.path.join(src, f"*{ext}")))

    for filename in filenames:
        dst_file = Path(dst) / Path(filename).name
        try:
            shutil.copy(filename, dst_file)
        except shutil.SameFileError:
            pass


def get_md5(text):
    if isinstance(text, str):
        text = text.encode()
    elif isinstance(text, (dict, list, tuple)):
        text = json.dumps(text, ensure_ascii=False).encode()

    return hashlib.md5(text).hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#


import logging
import logging.handlers
import os
from datetime import datetime
from datetime import timedelta
from datetime import timezone

logger = logging.getLogger(__name__)


FMT_LOG = "[%(asctime)s] %(levelname)s [%(funcName)s:%(lineno)d] %(message)s"
FMT_DATE = "%Y-%m-%d %H:%M:%S"


def time_to_beijing(*args):
    return datetime.utcnow().replace(tzinfo=timezone.utc).astimezone(timezone(timedelta(hours=8))).timetuple()


logging.Formatter.converter = time_to_beijing


def _configure_logger():
    level = os.environ.get("HAKKERO_LOG_LEVEL", "INFO")

    formatter = logging.Formatter(fmt=FMT_LOG, datefmt=FMT_DATE)

    handler_console = logging.StreamHandler()
    handler_console.setLevel(level)
    handler_console.setFormatter(formatter)
    logger.addHandler(handler_console)

    filename = os.environ.get("HAKKERO_LOG_FILE", None)
    if filename is not None:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        handler_file = logging.handlers.RotatingFileHandler(filename, maxBytes=5 * 1024 * 1024, backupCount=2)
        handler_file.setLevel(level)
        handler_file.setFormatter(formatter)
        logger.addHandler(handler_file)

    logger.propagate = False
    logger.setLevel(level)


_configure_logger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import re

import torch

from hakkero.dataset.image import process_messages
from hakkero.dataset.image import translate_messages
from hakkero.dataset.strategy.errors import TokenizationError
from hakkero.dataset.strategy.keep_thinks import change_chat_template_with_keep_think_remap
from hakkero.dataset.utils import IGNORE_INDEX


def legacy(data, tokenizer, **kwargs):
    context = "\n\n".join(
        [
            data[s].strip()
            for s in ("title", "summary", "abstract", "text", "question", "answer", "code")
            if s in data and data[s].strip()
        ]
    )

    target = data.get("label", "").strip()

    input = []
    label = []

    ids = tokenizer.encode(context, max_length=int(1e12), truncation=True)
    input.extend(ids)

    if target:
        if ids[-1] == tokenizer.eos_token_id:
            ids.pop()
        label.extend([IGNORE_INDEX for _ in ids])

        ids = tokenizer.encode(target, max_length=int(1e12), truncation=True)
        if ids[0] == tokenizer.bos_token_id:
            ids.pop(0)
        input.extend(ids)
        label.extend(ids)
    else:
        label.extend(ids)

    if len(input) <= 1:
        raise TokenizationError(
            "No valid keys in input, expect of: ('title', 'summary', 'abstract', 'text', 'question', 'answer', 'code')"
        )

    if kwargs.get("add_bos_token", False) and tokenizer.bos_token_id is not None:
        if input[0] != tokenizer.bos_token_id:
            input = [tokenizer.bos_token_id] + input
        if label[0] != tokenizer.bos_token_id:
            label = [tokenizer.bos_token_id] + label

    if kwargs.get("add_eos_token", False) and tokenizer.eos_token_id is not None:
        if input[-1] != tokenizer.eos_token_id:
            input = input + [tokenizer.eos_token_id]
        if label[-1] != tokenizer.eos_token_id:
            label = label + [tokenizer.eos_token_id]

    if kwargs.get("shift", True):
        return dict(input=torch.tensor(input[:-1], dtype=torch.long), label=torch.tensor(label[1:], dtype=torch.long))

    return dict(input=torch.tensor(input, dtype=torch.long), label=torch.tensor(label, dtype=torch.long))


def remove_ignore(content, ignore):
    if ignore is None:
        return content

    re_ignore = re.compile(f"(?:{ignore})$")
    return re_ignore.split(content)[0]


def _override_tokenizer(kwargs, tokenizer):
    old_chat_template = tokenizer.chat_template
    keep_think = kwargs.pop("keep_think", "no").lower()
    new_chat_template, keep_think = change_chat_template_with_keep_think_remap(old_chat_template, keep_think)
    tokenizer.chat_template = new_chat_template

    return keep_think, tokenizer, old_chat_template


# ----------------------------------------------------------------------------------------------------------------------
# keep_think: no - not keep <think>xx</think>, last - keep <think>xx</think> in last turn, all - keep <think>xx</think> in all turns
_tokenize_kwargs = {
    "padding": False,
    "truncation": False,
    "max_length": None,
    "add_special_tokens": False,
    "return_tensors": None,
}


def _process_single_turn(uid, is_last_turn, messages, tokenizer, keep_think, **kwargs):
    keep_think_in_context = keep_think == "all"
    keep_think_in_resp = keep_think == "all" or (keep_think == "last" and is_last_turn)

    aid = uid + 1

    text_context_ids = tokenizer.apply_chat_template(
        messages[:aid], add_generation_prompt=True, tokenize=False, keep_think=keep_think_in_context
    )
    text_context_ids = remove_ignore(text_context_ids, "<think>\n")
    context_ids = tokenizer(text_context_ids, **_tokenize_kwargs)["input_ids"]

    text_resp_ids_with_prefix = tokenizer.apply_chat_template(
        messages[: aid + 1], add_generation_prompt=False, tokenize=False, keep_think=keep_think_in_resp
    )

    text_prefix_ids = tokenizer.apply_chat_template(
        messages[:aid], add_generation_prompt=True, tokenize=False, keep_think=keep_think_in_resp
    )
    text_prefix_ids = remove_ignore(text_prefix_ids, "<think>\n")

    assert text_resp_ids_with_prefix[: len(text_prefix_ids)] == text_prefix_ids

    resp_ids = tokenizer(text_resp_ids_with_prefix[len(text_prefix_ids) :], **_tokenize_kwargs)["input_ids"]

    cur_input = context_ids + resp_ids
    cur_label = [IGNORE_INDEX for _ in context_ids] + resp_ids

    return cur_input, cur_label


# messages = [{"role": "user", "content": xxx}, {"role": "assistant", "content": xxx}, ...]
def huggingface_message(messages, tokenizer, **kwargs):
    keep_think, tokenizer, old_chat_template = _override_tokenizer(kwargs, tokenizer)

    n_turns = int((len(messages) if messages[0]["role"] == "user" else len(messages) - 1) / 2)

    input, label = [], []

    uid = next((i for i, v in enumerate(messages) if v["role"] == "user"), -1)
    cur_turn = 1

    while uid < len(messages):
        is_last_turn = cur_turn == n_turns

        cur_input, cur_label = _process_single_turn(uid, is_last_turn, messages, tokenizer, keep_think, **kwargs)

        input = input + cur_input[len(input) :]
        label = label + cur_label[len(label) :]

        uid += 2
        cur_turn += 1

    tokenizer.chat_template = old_chat_template

    if kwargs.get("shift", True):
        return dict(input=torch.tensor(input[:-1], dtype=torch.long), label=torch.tensor(label[1:], dtype=torch.long))

    return dict(input=torch.tensor(input, dtype=torch.long), label=torch.tensor(label, dtype=torch.long))


# data = {
#   "context": [
#       {"role": "user", "content": xxx},
#       {"role": "assistant", "content": xxx},
#       ...
#       {"role": "user", "content": xxx}
#   ],
#   "chosen": "xx",
#   "rejected": "xx"
# }
def huggingface_preference(data, tokenizer, **kwargs):
    keep_think, tokenizer, old_chat_template = _override_tokenizer(kwargs, tokenizer)

    keep_think_in_context = keep_think == "all"
    keep_think_in_resp = keep_think == "all" or keep_think == "last"  # keep think in chosen/rejected

    text_context_ids = tokenizer.apply_chat_template(
        data["context"], add_generation_prompt=True, tokenize=False, keep_think=keep_think_in_context
    )
    text_context_ids = remove_ignore(text_context_ids, "<think>\n")
    context_ids = tokenizer(text_context_ids, **_tokenize_kwargs)["input_ids"]

    text_prefix_ids = tokenizer.apply_chat_template(
        data["context"][-1:], add_generation_prompt=True, tokenize=False, keep_think=keep_think_in_resp
    )
    text_prefix_ids = remove_ignore(text_prefix_ids, "<think>\n")

    inputs = dict(chosen=[], rejected=[])
    labels = dict(chosen=[], rejected=[])

    for key in ("chosen", "rejected"):
        inputs[key].extend(context_ids)
        labels[key].extend(IGNORE_INDEX for _ in context_ids)

        text_response_ids_with_prefix = tokenizer.apply_chat_template(
            data["context"][-1:] + [{"role": "assistant", "content": data[key]}],
            add_generation_prompt=False,
            tokenize=False,
            keep_think=keep_think_in_resp,
        )

        assert text_response_ids_with_prefix[: len(text_prefix_ids)] == text_prefix_ids

        response_ids = tokenizer(text_response_ids_with_prefix[len(text_prefix_ids) :], **_tokenize_kwargs)["input_ids"]

        inputs[key].extend(response_ids)
        labels[key].extend(response_ids)

    tokenizer.chat_template = old_chat_template

    if kwargs.get("shift", True):
        return {
            "inputs": {key: torch.tensor(value[:-1]) for key, value in inputs.items()},
            "labels": {key: torch.tensor(value[1:]) for key, value in labels.items()},
        }

    return {
        "inputs": {key: torch.tensor(value) for key, value in inputs.items()},
        "labels": {key: torch.tensor(value) for key, value in labels.items()},
    }


# ----------------------------------------------------------------------------------------------------------------------
chatml_role = {
    "join": "\n",
    "user": "<|im_start|>user\n{}<|im_end|>",
    "system": "<|im_start|>system\n{}<|im_end|>",
    "assistant": "<|im_start|>assistant\n{}<|im_end|>",
    "assistant_start": "<|im_start|>assistant\n",
    "assistant_end": "<|im_end|>",
}


# messages = [{"role": "user", "content": xxx}, {"role": "assistant", "content": xxx}, ...]
def role_message(messages, tokenizer, template, context=None, **kwargs):
    assistant_start_ids = tokenizer.encode(
        template["assistant_start"], add_special_tokens=False, max_length=int(1e12), truncation=True
    )

    input, label, context = [], [], context
    for i, message in enumerate(messages, start=1):
        if message["role"] in ["system", "user"]:
            text = template[message["role"]].format(message["content"])
            context = text if context is None else template["join"].join([context, text])
        elif message["role"] == "assistant":
            # only tokenize and append context right before assistant message
            # context after assistant message is not useful
            context = template["join"].join([context, template["assistant_start"]])
            ids = tokenizer.encode(context, add_special_tokens=False, max_length=int(1e12), truncation=True)
            input.extend(ids)

            label.extend([IGNORE_INDEX for _ in ids])

            ids = tokenizer.encode(
                template["assistant_start"] + message["content"] + template["assistant_end"],
                add_special_tokens=False,
                max_length=int(1e12),
                truncation=True,
            )

            # a hack to avoid prepending space in the assistant response
            assert ids[: len(assistant_start_ids)] == assistant_start_ids
            input.extend(ids[len(assistant_start_ids) :])
            label.extend(ids[len(assistant_start_ids) :])
            context = ""
        else:
            raise ValueError(f"not supported role: {message['role']}")

    if kwargs.get("shift", True):
        return dict(input=torch.tensor(input[:-1]), label=torch.tensor(label[1:]))

    return dict(input=torch.tensor(input), label=torch.tensor(label))


def chatml_message(messages, tokenizer, **kwargs):
    return role_message(messages, tokenizer, chatml_role, **kwargs)


# data = {
#   "context": [{"role": "user", "content": xxx}, {"role": "assistant", "content": xxx}, ...],
#   "chosen": "xx",
#   "rejected": "xx"
# }
def role_preference(data, tokenizer, template, **kwargs):
    assistant_start_ids = tokenizer.encode(
        template["assistant_start"], add_special_tokens=False, max_length=int(1e12), truncation=True
    )
    inputs = dict(chosen=[], rejected=[])
    labels = dict(chosen=[], rejected=[])

    context = template["join"].join(
        [template[message["role"]].format(message["content"]) for message in data["context"]]
        + [template["assistant_start"]]
    )
    context_ids = tokenizer.encode(context, add_special_tokens=False, max_length=int(1e12), truncation=True)

    for key in ("chosen", "rejected"):
        inputs[key].extend(context_ids)
        labels[key].extend(IGNORE_INDEX for _ in context_ids)
        response_ids_with_prefix = tokenizer.encode(
            template["assistant_start"] + data[key] + template["assistant_end"],
            add_special_tokens=False,
            max_length=int(1e12),
            truncation=True,
        )
        assert response_ids_with_prefix[: len(assistant_start_ids)] == assistant_start_ids
        response_ids = response_ids_with_prefix[len(assistant_start_ids) :]
        inputs[key].extend(response_ids)
        labels[key].extend(response_ids)

    if kwargs.get("shift", True):
        return {
            "inputs": {key: torch.tensor(value[:-1]) for key, value in inputs.items()},
            "labels": {key: torch.tensor(value[1:]) for key, value in labels.items()},
        }

    return {
        "inputs": {key: torch.tensor(value) for key, value in inputs.items()},
        "labels": {key: torch.tensor(value) for key, value in labels.items()},
    }


def chatml_preference(data, tokenizer, **kwargs):
    return role_preference(data, tokenizer, chatml_role, **kwargs)


# ----------------------------------------------------------------------------------------------------------------------
# for MM LLM

qwen2_system = "You are a helpful assistant."


def chatml_qwen2_vl_message(messages, tokenizer, processor, path, **kwargs):
    messages, images = translate_messages(messages, path)
    mm_inputs = None
    if len(images) > 0:
        messages, mm_inputs = process_messages(messages, images, processor)

    msg = role_message(messages, tokenizer, chatml_role, context=chatml_role["system"].format(qwen2_system), **kwargs)
    if mm_inputs is not None:
        msg.update(mm_inputs)

    return msg
